/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.check;

import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;
import java.util.concurrent.atomic.AtomicInteger;

import org.citygml4j.factory.GMLGeometryFactory;
import org.citygml4j.model.citygml.core.CityModel;

import de.hft.stuttgart.citydoctor2.check.error.SchematronError;
import de.hft.stuttgart.citydoctor2.checks.SvrlContentHandler;
import de.hft.stuttgart.citydoctor2.checks.util.FeatureCheckedListener;
import de.hft.stuttgart.citydoctor2.datastructure.BridgeObject;
import de.hft.stuttgart.citydoctor2.datastructure.Building;
import de.hft.stuttgart.citydoctor2.datastructure.CityObject;
import de.hft.stuttgart.citydoctor2.datastructure.LandObject;
import de.hft.stuttgart.citydoctor2.datastructure.TransportationObject;
import de.hft.stuttgart.citydoctor2.datastructure.Vegetation;
import de.hft.stuttgart.citydoctor2.datastructure.WaterObject;
import de.hft.stuttgart.citydoctor2.parser.CityGmlConsumer;
import de.hft.stuttgart.citydoctor2.reporting.XmlStreamReporter;
import de.hft.stuttgart.citydoctor2.reporting.pdf.PdfStreamReporter;
import de.hft.stuttgart.citydoctor2.utils.Localization;
import de.hft.stuttgart.citydoctor2.utils.QualityADEUtils;
import de.hft.stuttgart.quality.model.Validation;
import de.hft.stuttgart.quality.model.jaxb.ErrorStatistics;
import de.hft.stuttgart.quality.model.jaxb.FeatureStatistics;
import de.hft.stuttgart.quality.model.jaxb.Statistics;

public class StreamCityGmlConsumer implements CityGmlConsumer {

	private Checker c;
	private XmlStreamReporter xmlReporter;
	private PdfStreamReporter pdfReporter;
	private SvrlContentHandler handler;
	private Map<ErrorId, AtomicInteger> errorCount;
	private GMLGeometryFactory gmlFactory;
	private ValidationConfiguration config;
	private Statistics statistics;
	private FeatureStatistics buildingStatistics;
	private FeatureStatistics bridgeStatistics;
	private FeatureStatistics transportationStatistics;
	private FeatureStatistics vegetationStatistics;
	private FeatureStatistics landStatistics;
	private FeatureStatistics waterStatistics;
	private Validation val;
	private FeatureCheckedListener l;

	public StreamCityGmlConsumer(Checker c, XmlStreamReporter xmlReporter, PdfStreamReporter pdfReporter,
			SvrlContentHandler handler, ValidationConfiguration config, FeatureCheckedListener l) {
		this.c = c;
		this.xmlReporter = xmlReporter;
		this.pdfReporter = pdfReporter;
		this.handler = handler;
		this.config = config;
		this.l = l;
		errorCount = new HashMap<>();
		gmlFactory = new GMLGeometryFactory();

		val = new Validation();
		val.setValidationDate(ZonedDateTime.now());
		val.setValidationSoftware("CityDoctor " + Localization.getText(Localization.VERSION));
		statistics = new Statistics();
		buildingStatistics = new FeatureStatistics();
		statistics.setNumErrorBuildings(buildingStatistics);
		bridgeStatistics = new FeatureStatistics();
		statistics.setNumErrorBridgeObjects(bridgeStatistics);
		transportationStatistics = new FeatureStatistics();
		statistics.setNumErrorTransportation(transportationStatistics);
		vegetationStatistics = new FeatureStatistics();
		statistics.setNumErrorVegetation(vegetationStatistics);
		landStatistics = new FeatureStatistics();
		statistics.setNumErrorLandObjects(landStatistics);
		waterStatistics = new FeatureStatistics();
		statistics.setNumErrorWaterObjects(waterStatistics);

	}

	@Override
	public void accept(CityObject co) {
		c.checkFeature(xmlReporter, pdfReporter, co);

		if (handler != null) {
			List<SchematronError> errors = handler.getFeatureErrors().get(co.getGmlId().getGmlString());
			if (errors != null) {
				Checker.handleSchematronErrorsForCityObject(errors, co);
			}
		}

		// remove existing quality ade datastructure if existing
		QualityADEUtils.removeValidationResult(co);
		// store quality ade datastructures in cityobject
		QualityADEUtils.writeQualityAde(co);
		// recreate geometry
		co.reCreateGeometries(gmlFactory, config.getParserConfiguration());

		// store result in statistics
		applyToStatistics(buildingStatistics, bridgeStatistics, transportationStatistics, vegetationStatistics,
				landStatistics, waterStatistics, co);

		// add errors to statistics
		List<CheckError> errorList = new ArrayList<>();
		co.collectContainedErrors(errorList);
		Set<CheckError> errors = new HashSet<>(errorList);
		for (CheckError e : errors) {
			errorCount.compute(e.getErrorId(), (k, v) -> {
				if (v == null) {
					return new AtomicInteger(1);
				}
				v.incrementAndGet();
				return v;
			});
		}

		if (l != null) {
			l.featureChecked(co);
		}
	}

	@Override
	public void accept(CityModel cm) {
		QualityADEUtils.removeValidation(cm);
		for (Entry<ErrorId, AtomicInteger> e : errorCount.entrySet()) {
			ErrorStatistics stats = new ErrorStatistics();
			stats.setAmount(e.getValue().get());
			stats.setName(QualityADEUtils.mapErrorIdToAdeId(e.getKey()));
			statistics.getErrorStatistics().add(stats);
		}
		val.setStatistics(statistics);
		val.setValidationPlan(c.createValidationPlan());

		cm.addGenericApplicationPropertyOfCityModel(val);
	}

	private static void applyToStatistics(FeatureStatistics buildingStatistics, FeatureStatistics bridgeStatistics,
			FeatureStatistics transportationStatistics, FeatureStatistics vegetationStatistics,
			FeatureStatistics landStatistics, FeatureStatistics waterStatistics, CityObject co) {
		if (co.isValidated()) {
			if (co instanceof Building) {
				countForFeatureStatistics(buildingStatistics, co);
			} else if (co instanceof TransportationObject) {
				countForFeatureStatistics(transportationStatistics, co);
			} else if (co instanceof BridgeObject) {
				countForFeatureStatistics(bridgeStatistics, co);
			} else if (co instanceof WaterObject) {
				countForFeatureStatistics(waterStatistics, co);
			} else if (co instanceof LandObject) {
				countForFeatureStatistics(landStatistics, co);
			} else if (co instanceof Vegetation) {
				countForFeatureStatistics(vegetationStatistics, co);
			}
		}
	}

	private static void countForFeatureStatistics(FeatureStatistics featureStatistics, CityObject co) {
		featureStatistics.setNumChecked(featureStatistics.getNumChecked() + 1);
		if (co.containsAnyError()) {
			featureStatistics.setNumErrors(featureStatistics.getNumErrors() + 1);
		}
	}

}
